# Probabilistic Graphical Model Library
#
# Copyright (C) 2016
#
# author: Natapol Pornputtapong <natapol.por@gmail.com>
#
# Documentation: Natapol Pornputtapong
#

module PGM

  # describe bidirectionally function to Graph
  module Bidirectional

    # Instantiate DiscreateVariable
    #
    # @param edgelist_class [Object] Object to store edge class
    # @param other_graphs [PGM::BayesianNet] List of variable states
    def initialize(edgelist_class = Set, *other_graphs)
      @vertices_rev_dict = Hash.new
      @variables = Hash.new
      super
    end

    # Copy internal vertices_dict
    #
    # @param orig [PGM::BayesianNet] Copy from other graph
    def initialize_copy(orig)
      super
      @vertices_rev_dict = orig.instance_eval { @vertices_rev_dict }.dup
      @vertices_rev_dict.keys.each do |v|
        @vertices_rev_dict[v] = @vertices_rev_dict[v].dup
      end
      @variables = orig.instance_eval { @variables }.dup
    end

    # Return variable with the name
    #
    # @param name [String, Symbol] The name of variable
    # @return [PGM::Variable] Variable object
    def var(name=nil)
      if nil
        return @variables
      else
        return @variables[name.to_sym]
      end
    end

    # Create variable
    #
    # @param name [String, Symbol] Name of variable
    # @param states [Array<String, Symbol>] List of variable states
    # @param parents [Array<PGM::Variable>] List of parent variable
    # @param type [PGM::Variable] type of variable
    def create_var(name, states, parents: [], type: PGM::DiscreateVariable)
      add_vertex(name)
      @variables[name.to_sym] = type.new(self, name, states, parents: parents)
    end

    # Check if model has the variable named as input
    #
    # @param name [String, Symbol] Name of variable
    # @return [Boolean] if model has the variable
    def has_var?(name)
      return @variables.has_key?(name.to_sym)
    end

    # assign link from u to v
    #
    # @param u [PGM::Variable] Variable to be linked
    # @param v [PGM::Variable] Variable to be linked
    def create_link(u, v)
      if has_var?(u) && has_var?(v)
        var(v).to(var(u))
      end
    end

    protected

    def basic_add_edge(u, v)
      super
      @vertices_rev_dict[v].add(u)
    end

    # See MutableGraph#add_vertex.
    #
    # If the vertex is already in the graph (using eql?), the method does
    # nothing.
    #
    def add_vertex(v)
      super
      @vertices_rev_dict[v] ||= @edgelist_class.new
    end

    # See MutableGraph#remove_vertex.
    #
    def remove_vertex(v)
      super
      @vertices_rev_dict.delete(v)
      # remove v from all adjacency lists
      @vertices_rev_dict.each_value { |adjList| adjList.delete(v) }
    end

    # See MutableGraph::remove_edge.
    #
    # Also removes (v,u)
    #
    def remove_edge(u, v)
      super
      @vertices_rev_dict[v].delete(u) unless @vertices_dict[v].nil?
    end

    def each_parent(v, &b) # :nodoc:
      parent_list = (@vertices_rev_dict[v] or raise NoVertexError, "No vertex #{v}.")
      parent_list.each(&b)
    end

    def parent_vertices(v)
      r = []
      each_parent(v) { |u| r << u }
      r
    end

  end

end
