# Probabilistic Graphical Model Library
#
# Copyright (C) 2016
#
# author: Natapol Pornputtapong <natapol.por@gmail.com>
#
# Documentation: Natapol Pornputtapong
#

require 'rgl/adjacency'
require 'rgl/topsort'
require 'rgl/dot'

require 'pgm/variable'
require 'pgm/bidirectional'

module PGM
  # Bayesian Network Class
  #
  # Example:
  #
  # create model
  #   bayenet = PGM::BayesianNet.new()
  #   bayenet.create_var(:cloudy, [:t, :f])
  #   bayenet.create_var(:sprinkler, [:t, :f])
  #   bayenet.create_var(:rain, [:t, :f])
  #   bayenet.create_var(:grass_wet, [:t, :f])
  #   bayenet.create_link(:cloudy, :sprinkler)
  #   bayenet.create_link(:cloudy, :rain)
  #   bayenet.create_link(:sprinkler, :grass_wet)
  #   bayenet.create_link(:rain, :grass_wet)
  #
  # learn with data
  #   bayenet.learn([
  #     {:cloudy => :t, :sprinkler => :f, :rain => :t, :grass_wet => :t},
  #     {:cloudy => :t, :sprinkler => :t, :rain => :f, :grass_wet => :t},
  #     {:cloudy => :f, :sprinkler => :f, :rain => :t, :grass_wet => :t},
  #     {:cloudy => :t, :sprinkler => :f, :rain => :t, :grass_wet => :t},
  #     {:cloudy => :f, :sprinkler => :t, :rain => :f, :grass_wet => :f},
  #     {:cloudy => :f, :sprinkler => :f, :rain => :f, :grass_wet => :f},
  #     {:cloudy => :f, :sprinkler => :f, :rain => :f, :grass_wet => :f},
  #     {:cloudy => :t, :sprinkler => :f, :rain => :t, :grass_wet => :t},
  #     {:cloudy => :t, :sprinkler => :f, :rain => :f, :grass_wet => :f},
  #     {:cloudy => :f, :sprinkler => :f, :rain => :f, :grass_wet => :f},
  #   ])
  #
  # print out the conditional_probability_table
  #   bayenet.puts_conditional_probability_table
  #
  # create graph from model
  #   bayenet.write_to_graphic_file('jpg')
  class BayesianNet < RGL::DirectedAdjacencyGraph

    include PGM::Bidirectional

    # Find common parents of input nodes
    #
    # @param vs [Array<PGM::Variable>] List of vertices
    # @return [Array<PGM::Variable>] List parent vertices
    def common_parents(*vs)
      r = nil
      vs.each do |v|
        if r
          r &= parent_vertices(v)
        else
          r = parent_vertices(v)
        end
      end
      r ||= []
    end

    # Check if there are common parents of input vertices
    #
    # @param vs [Array<PGM::Variable>] List of variables
    # @return [Boolean] Is there common parents
    def common_parents?(*vs)
      return !common_parents(*vs).empty?
    end

    # Find common children of input vertices
    #
    # @param vs [Array<PGM::Variable>] List of vertices
    # @return [Array<PGM::Variable>] List child vertices
    def common_children(*vs)
      r = nil
      vs.each do |v|
        if r
          r &= adjacent_vertices(v)
        else
          r = adjacent_vertices(v)
        end
      end
      r ||= []
    end

    # Check if there are common children of input vertices
    #
    # @param vs [Array<PGM::Variable>] List of vertices
    # @return [Boolean] Is there common children
    def common_children?(*vs)
      return !common_children(*vs).empty?
    end

    # Check if vertices are in cascading topology
    #
    # @param vs [Array<PGM::Variable>] List of vertices
    # @return [Boolean] Are vertices in cascading topology
    def cascade?(*vs)
      prev_children = nil
      r = true
      if vs.length > 2
        vs.each do |v|
          prev_children ||= [v]
          if !prev_children.include?(v)
            r = false
            break
          else
            prev_children = adjacent_vertices(v)
          end
        end
      else
        r = false
      end
      return r
    end

    # Check if vertices are in V-Structure
    #
    # @param vs [Array<PGM::Variable>] List of vertices
    # @return [Boolean] Are vertices in V-Structure
    def v_tructure?(*vs)
      if vs.length == 3
        return parent_vertices(vs[0]).sort == vs[1..-1].sort
      else
        return false
      end
    end

    # Generate string notation of Baysian Network
    #
    # @return [String] Model notation
    def model_notation
      line = []
      each_vertex do |v|
        parents = parent_vertices(v)
        line.push("P(#{v}#{parents.empty? ? '' : "|#{parents.join(',')}"})")
      end
      return "P(#{vertices.sort.join(',')}) = #{line.sort.join('')}"
    end

    # learn model from array of condition and automatically calculate the
    # conditional probability table. Condition is in a Hash format { var_name1: state_name1, var_name2: state_name2 }
    #
    # @param arr [Array<Hash{Symbol => Symbol,String}>] Array of data for learning in Hash objects
    def learn(arr)
      @variables.each_value do |var|
        var.learn(arr)
      end
    end

    # Learn the model with discrete data.
    # Condition is in a Hash format { var_name1: state_name1, var_name2: state_name2 }
    #
    # @param condition [Hash{Symbol => Symbol,String}] Data for learning in a Hash objects
    def add_data(condition)
      @variables.each_value do |var|
        var.add_data(condition)
      end
    end

    # Set conditional probability to a random variable
    # Condition is in a Hash format { var_name1: state_name1, var_name2: state_name2 }
    #
    # @param varname [String, Symbol] Random variable name
    # @param prob [Float] Probability
    # @param condition [Hash{Symbol => Symbol,String}] Data for learning in Hash objects
    def set_conditional_probability(varname, prob, condition)
      var(varname.to_sym).set_conditional_probability(prob, condition)
    end

    # Calculate conditional probability table from learned data
    #
    def calculate_conditional_probability_table
      @variables.each_value do |var|
        var.calculate_conditional_probability_table
      end
    end

    alias_method :cal_cpt, :calculate_conditional_probability_table

    # puts conditional probability table on the screen
    #
    # @param cellsize [Integer] Size of each cell
    def puts_conditional_probability_table(cellsize = 25)
      @variables.each_value do |val|
        val.puts_conditional_probability_table(cellsize)
      end
    end

    alias_method :puts_cpt, :puts_conditional_probability_table

    # Order random variable by ancestral order
    #
    def ancestral_ordering
      return self.topsort_iterator.to_a
    end
  end

end
