# Probabilistic Graphical Model Library
#
# Copyright (C) 2016
#
# author: Natapol Pornputtapong <natapol.por@gmail.com>
#
# Documentation: Natapol Pornputtapong
#

require 'narray'

module PGM

  # Most ancestral class for Random variable
  class Variable
    def initialize
      raise NotImplementedError
    end
  end

  # A class for discrete random variable
  class DiscreateVariable < Variable

    attr_reader :name, :model, :states, :parents, :children

    # Instantiate DiscreateVariable
    #
    # @param model [PGM::BayesianNet] Model for variable to be added
    # @param name [String, Symbol] Name of variable
    # @param states [Array<String, Symbol>] List of variable states
    # @param parents [Array<PGM::Variable>] List of parent vertices
    # @param children [Array<PGM::Variable>] List of child vertices
    def initialize(model, name, states, parents: [], children: [])
      @name = name.to_sym
      @model = model
      @states = states.map{|x| x.to_sym}.sort
      @parents = parents
      @children = children
      self.create_conditional_probability_table
    end

    # compare variables
    #
    # @param other [PGM::Variable] Model for variable to be added
    # @return [Integer] List of child vertices
    def <=>(other)
      if other.is_a?(PGM::Variable)
        if @model == other.model
          return @name <=> other.name
        else
          return @model <=> other.model
        end
      else
        raise TypeError, "#{other} is not PGM::Variable"
      end
    end

    # assign link from self to other
    #
    # @param other [PGM::Variable] Variable to be linked
    def to(other)
      other.add_parent(self)
      self.add_child(other)
      @model.add_edge(@name, other.name)
    end

    # learn model from array of condition and automatically calculate the
    # conditional probability table
    #
    # @param arr [Array<Hash{Symbol => Symbol,String}>] Array of data for learning in Hash objects
    def learn(arr)
      arr.each do |item|
        self.add_data(item)
      end
      self.calculate_conditional_probability_table
    end

    # learn model from data.
    # Condition is in a Hash format { var_name1: state_name1, var_name2: state_name2 }
    #
    # @param condition [Hash{Symbol => Symbol,String}] Data for learning in Hash objects
    def add_data(condition)
      @count_table[*self.data_coor(condition)] += 1
    end

    # calculate conditional probability table
    def calculate_conditional_probability_table
      @probability_table = @count_table.floor / @count_table.sum(1).to_f
    end

    alias_method :cal_cpt, :calculate_conditional_probability_table

    # puts conditional probability table on the screen
    #
    # @param cellsize [Integer] Size of each cell
    def puts_conditional_probability_table(cellsize = 25)
      puts "#{@name.to_s.center(cellsize)}|#{@adjoint_var_list.map{|x| "P(#{x})".center(cellsize)}.join('|')}"
      0.upto(@states.length - 1).each do |ind|
        puts "#{@states[ind].to_s.rjust(cellsize)}|#{@probability_table[true, ind].to_a.map{|x| x.round(5).to_s.rjust(cellsize)}.join('|')}"
      end
    end

    alias_method :puts_cpt, :puts_conditional_probability_table

    # assign conditional probability for specified condition.
    # Condition is in a Hash format { var_name1: state_name1, var_name2: state_name2 }
    #
    # @param prob [Float] probability
    # @param condition [Hash{Symbol => Symbol,String}] condition in Hash objects
    def set_conditional_probability(prob, condition)
      @probability_table[*self.data_coor(condition)] = prob
    end

    alias_method :set_cp, :set_conditional_probability

    # return conditional probability of the specified condition.
    # Condition is in a Hash format { var_name1: state_name1, var_name2: state_name2 }
    #
    # @param condition [Hash{Symbol => Symbol,String}] condition in Hash objects
    # @return [Float] probability
    def conditional_probability(condition)
      return @probability_table[*self.data_coor(condition)]
    end

    protected

    def data_coor(condition)
      #[column, row]
      notation = []
      @parents.map{|x| x.name.to_sym}.each do |name|
        notation.push("#{name}=#{condition[name]}")
      end
      return [
                @adjoint_var_list.include?(notation.join(',')) ? @adjoint_var_list.index(notation.join(',')) : true,
                @states.include?(condition[@name]) ? @states.index(condition[@name]) : true
              ]
    end

    def state_notation
      notations = []
      self.states.each do |state|
        notations.push("#{name}=#{state}")
      end
      return notations
    end

    def create_conditional_probability_table
      list = []
      @parents.each do |parent|
        list.push(parent.state_notation)
      end
      if list.length == 0
        @adjoint_var_list = [@name]
      elsif list.length == 1
        @adjoint_var_list = list[0]
      else
        @adjoint_var_list = list[0].product(*list[1..-1]).map{|x| x.join(',')}
      end
      @count_table = NArray.sfloat(@adjoint_var_list.length, @states.length).fill(0.0000001)
      @probability_table = NArray.sfloat(@adjoint_var_list.length, @states.length)
    end

    def add_parent(var)
      @parents.push(var)
      @parents.sort!
      self.create_conditional_probability_table
    end

    def add_child(var)
      @children.push(var)
      @children.sort!
    end

  end

  # A class for concrete random variable
  class ConcreteVariable < Variable

  end
end
